@@ -1,21 +1,33 @@
 Bclust <- function(data, method.d="euclidean", method.c="ward.D",
-FUN=function(.x) hclust(dist(.x, method=method.d), method=method.c),
-iter=1000, mc.cores=1, monitor=TRUE, bootstrap=TRUE) {
+ FUN=function(.x) hclust(dist(.x, method=method.d), method=method.c),
+ iter=1000, mc.cores=1, monitor=TRUE, bootstrap=TRUE, relative=FALSE,
+ hclist=NULL) {
  .calc.matches <- function(origin, current, nc=ncol(origin)) {
   one <- tcrossprod(origin, current) # both 1
   zero <- tcrossprod(1 - origin, 1 - current) # both 0
-  return(rowSums(one + zero == nc)) # calc matches
+  res <- rowSums(one + zero == nc) # absolute matches (cluster present or absent)
+  zeros <- sum(res == 0) # how many non-matches
+  if (zeros > 0 && relative) { # select non-matches and calculate _relative matches_
+  dsts <- as.matrix(dist(rbind(origin[res == 0, ], current), # find distances
+   method="binary"))[seq_len(zeros), -seq_len(zeros), drop=FALSE] # asymmetric binary because "0" and "1" are principally different
+  res[res == 0] <- apply(dsts, 1, min) # replace zeroes with closest distances
+  }
+  res
  }
  fun <- match.fun(FUN)
- hcl <- fun(data)
+ if (is.null(hclist)) hcl <- fun(data) else hcl <- hclist[[1]]
  origin <- Hcl2mat(hcl)
  if (!bootstrap) iter <- ncol(data) # jacknife
- v <- parallel::mclapply(seq_len(iter), function(y, origin, size, fun, nco) {
+ if (!is.null(hclist)) iter <- length(hclist) - 1 # pre-built list
+ v <- parallel::mclapply(seq_len(iter), function(.y, origin, size, fun, nco) {
+  if (!is.null(hclist)) {
+   current <- Hcl2mat(hclist[[.y + 1]]) # pre-built list
+  } else {
   if(!bootstrap) {
-   current <- Hcl2mat(fun(data[, -y])) # jacknife
+   current <- Hcl2mat(fun(data[, -.y])) # jacknife
   } else {
    current <- Hcl2mat(fun(data[, sample.int(ncol(data), replace=TRUE)]))
-  }
+  }}
   if (monitor) cat(".")
   return(list(matches=.calc.matches(origin, current, nco), current=current))
   },
@@ -23,8 +35,11 @@ iter=1000, mc.cores=1, monitor=TRUE, bootstrap=TRUE) {
  if (monitor) cat("\n")
  values <- colSums(do.call(rbind, lapply(v, `[[`, "matches")))/iter
  con <- t(Reduce("+", lapply(v, `[[`, "current")))/iter
- row.names(con) <- row.names(data)
- return(list(values=values, hclust=hcl, consensus=con))
+ if (is.null(hclist)) row.names(con) <- row.names(data) else row.names(con) <- hcl$labels
+ b.meth <- ifelse(bootstrap, "Bootstrap", "Jackknife")
+ b.clust <- list(values=values, hclust=hcl, consensus=con, meth=b.meth, iter=iter)
+ class(b.clust) <- "Bclust"
+ b.clust
 }
 
 ## ===
@@ -121,3 +136,11 @@ if (method == "points") {
  }
 invisible(list(coords=coords, labels=values))
 }
+
+# ===
+
+plot.Bclust <- function (x, main="", xlab=NULL, ...) {
+ if (is.null(xlab)) xlab <- paste(x$meth, ", ", x$iter, " replicates", sep = "")
+ plot(x$hclust, main=main, xlab=xlab, ...)
+ Bclabels(x$hclust, x$values, percent=TRUE, pos=3, offset=0.1)
+}
